@@ -1,4 +1,8 @@
-use crate::{failure, failure::Failure, toastfile::Task};
+use crate::{
+    failure,
+    failure::Failure,
+    toastfile::{command, location, user, Task, Toastfile},
+};
 use sha2::{Digest, Sha256};
 use std::{
     collections::HashMap,
@@ -86,6 +90,7 @@ pub fn hash_read<R: Read>(input: &mut R) -> Result<String, Failure> {
 pub fn image_name(
     previous_image: &str,
     docker_repo: &str,
+    toastfile: &Toastfile,
     task: &Task,
     input_files_hash: &str,
     environment: &HashMap<String, String>,
@@ -119,13 +124,13 @@ pub fn image_name(
     cache_key = combine(&cache_key, input_files_hash);
 
     // Incorporate the location.
-    cache_key = combine(&cache_key, &task.location);
+    cache_key = combine(&cache_key, &location(toastfile, task));
 
     // Incorporate the user.
-    cache_key = combine(&cache_key, &task.user);
+    cache_key = combine(&cache_key, &user(toastfile, task));
 
     // Incorporate the command.
-    cache_key = combine(&cache_key, &task.command);
+    cache_key = combine(&cache_key, &command(toastfile, task));
 
     // We add this "toast-" prefix because Docker has a rule that tags cannot be 64-byte hexadecimal
     // strings. See this for more details: https://github.com/moby/moby/issues/20972
@@ -136,10 +141,39 @@ pub fn image_name(
 mod tests {
     use crate::{
         cache::{combine, hash_read, image_name, CryptoHash},
-        toastfile::{Task, DEFAULT_LOCATION, DEFAULT_USER},
+        toastfile::{Task, Toastfile, DEFAULT_LOCATION, DEFAULT_USER},
     };
     use std::{collections::HashMap, path::Path};
 
+    fn toastfile_with_task(foo_task: Task) -> Toastfile {
+        let mut tasks = HashMap::new();
+        tasks.insert("foo".to_owned(), foo_task);
+
+        Toastfile {
+            image: "encom:os-12".to_owned(),
+            default: None,
+            location: Path::new(DEFAULT_LOCATION).to_owned(),
+            user: DEFAULT_USER.to_owned(),
+            command_prefix: "".to_owned(),
+            tasks,
+        }
+    }
+
+    fn toastfile_with_tasks(foo_task: Task, bar_task: Task) -> Toastfile {
+        let mut tasks = HashMap::new();
+        tasks.insert("foo".to_owned(), foo_task);
+        tasks.insert("bar".to_owned(), bar_task);
+
+        Toastfile {
+            image: "encom:os-12".to_owned(),
+            default: None,
+            location: Path::new(DEFAULT_LOCATION).to_owned(),
+            user: DEFAULT_USER.to_owned(),
+            command_prefix: "".to_owned(),
+            tasks,
+        }
+    }
+
     #[test]
     fn hash_str_pure() {
         assert_eq!("foo".crypto_hash(), "foo".crypto_hash());
@@ -219,11 +253,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: String::new(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_task(task);
+
         let input_files_hash = "grault";
 
         let full_environment = HashMap::new();
@@ -233,7 +271,8 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
@@ -260,11 +299,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_task(task);
+
         let input_files_hash = "grault";
 
         let mut full_environment = HashMap::new();
@@ -274,14 +317,16 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
             image_name(
                 previous_image,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
@@ -306,11 +351,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_task(task);
+
         let input_files_hash = "grault";
 
         let full_environment = HashMap::new();
@@ -319,14 +368,16 @@ mod tests {
             image_name(
                 previous_image1,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
             image_name(
                 previous_image2,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
@@ -358,9 +409,11 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
         let task2 = Task {
@@ -375,11 +428,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_tasks(task1, task2);
+
         let input_files_hash = "grault";
 
         let mut full_environment = HashMap::new();
@@ -390,14 +447,16 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task1,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
             image_name(
                 previous_image,
                 docker_repo,
-                &task2,
+                &toastfile,
+                &toastfile.tasks["bar"],
                 input_files_hash,
                 &full_environment,
             ),
@@ -427,9 +486,11 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
         let task2 = Task {
@@ -444,11 +505,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_tasks(task1, task2);
+
         let input_files_hash = "grault";
 
         let mut full_environment = HashMap::new();
@@ -459,14 +524,16 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task1,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
             image_name(
                 previous_image,
                 docker_repo,
-                &task2,
+                &toastfile,
+                &toastfile.tasks["bar"],
                 input_files_hash,
                 &full_environment,
             ),
@@ -493,11 +560,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_task(task);
+
         let input_files_hash = "grault";
 
         let mut full_environment1 = HashMap::new();
@@ -509,14 +580,16 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment1,
             ),
             image_name(
                 previous_image,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment2,
             ),
@@ -540,11 +613,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_task(task);
+
         let input_files_hash1 = "foo";
         let input_files_hash2 = "bar";
 
@@ -554,14 +631,16 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash1,
                 &full_environment,
             ),
             image_name(
                 previous_image,
                 docker_repo,
-                &task,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash2,
                 &full_environment,
             ),
@@ -585,9 +664,11 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new("/foo").to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: Some(Path::new("/foo").to_owned()),
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
         let task2 = Task {
@@ -602,11 +683,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new("/bar").to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: Some(Path::new("/bar").to_owned()),
+            user: None,
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_tasks(task1, task2);
+
         let input_files_hash = "grault";
 
         let full_environment = HashMap::new();
@@ -615,14 +700,16 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task1,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
             image_name(
                 previous_image,
                 docker_repo,
-                &task2,
+                &toastfile,
+                &toastfile.tasks["bar"],
                 input_files_hash,
                 &full_environment,
             ),
@@ -646,9 +733,11 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: "foo".to_owned(),
+            location: None,
+            user: Some("foo".to_owned()),
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
         let task2 = Task {
@@ -663,11 +752,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: "bar".to_owned(),
+            location: None,
+            user: Some("bar".to_owned()),
             command: "echo wibble".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_tasks(task1, task2);
+
         let input_files_hash = "grault";
 
         let full_environment = HashMap::new();
@@ -676,14 +769,16 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task1,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
             image_name(
                 previous_image,
                 docker_repo,
-                &task2,
+                &toastfile,
+                &toastfile.tasks["bar"],
                 input_files_hash,
                 &full_environment,
             ),
@@ -707,9 +802,11 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo foo".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
         let task2 = Task {
@@ -724,11 +821,15 @@ mod tests {
             mount_paths: vec![],
             mount_readonly: false,
             ports: vec![],
-            location: Path::new(DEFAULT_LOCATION).to_owned(),
-            user: DEFAULT_USER.to_owned(),
+            location: None,
+            user: None,
             command: "echo bar".to_owned(),
+            command_prefix: None,
+            extra_docker_arguments: vec![],
         };
 
+        let toastfile = toastfile_with_tasks(task1, task2);
+
         let input_files_hash = "grault";
 
         let full_environment = HashMap::new();
@@ -737,14 +838,16 @@ mod tests {
             image_name(
                 previous_image,
                 docker_repo,
-                &task1,
+                &toastfile,
+                &toastfile.tasks["foo"],
                 input_files_hash,
                 &full_environment,
             ),
             image_name(
                 previous_image,
                 docker_repo,
-                &task2,
+                &toastfile,
+                &toastfile.tasks["bar"],
                 input_files_hash,
                 &full_environment,
             ),
